"""_____________________________________________________________________

:PROJECT: ChemDataReader

* chemdatareader implementation *

:details:  chemdatareader implementation.

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

import csv
import asyncio
from chemdatareader.chemdatareader_logger import get_logger

import chemdatareader.chemdatacollector_builder_interface as cdbi
import chemdatareader.collectors.pubchem_collector

import chemdatareader.chemdatareader_interface as cdri


class ChemDatReader(cdri.ChemDatReaderInterface):
    """Implementation of the ChemDatReader Interface"""

    def __init__(self) -> None:
        self.logger = get_logger(__name__)

        self.substance_name_list = []
        self.pubchem_service = cdbi.services.get("pubchem", pubchem_client_key="key", pubchem_client_secret="secret")

    def test_connections(self) -> str:
        """connection test module

        :param name: person to greet
        :type name: str
        """
        self.logger.debug(f"connection to : {self.pubchem_service.test_connection()}")
        return self.pubchem_service.test_connection()
    
    async def get_compounds_by_name(self, name_list: list[str] = [], operation: str = "record", output_format: str = "JSON"):
        """
        get all compounds of a list of names or CAS numbers
        """
        self.compound_list = []
        self.operation = operation
        self.output_format = output_format

        await self.pubchem_service.get_compounds_by_name(name_list, operation, output_format)
    
        return self.pubchem_service.compound_list

    async def read_substance_names_csv(self, csv_filename: str = None, column_name: str = None) -> None:
        """csv reader module
        :param: csv_filename: filename of the csv file
        :type: csv_filename: str
        """

        self.logger.debug(f"csv_filename : {csv_filename}")
        try:
            with open(csv_filename, newline="") as csvfile:
                reader = csv.DictReader(csvfile)
                for row in reader:
                    self.logger.debug(f"row {row}")
                    self.logger.debug(f"{column_name}: {row[column_name] }")

                    self.substance_name_list.append(row[column_name])

            self.logger.debug(f"namelist: {self.substance_name_list}")
            await self.pubchem_service.get_compounds_by_name(self.substance_name_list)

            #self.logger.debug(f"compound_list: {compound_list}")

            #return compound_list

        except FileNotFoundError:
            self.logger.error(f"csv file {csv_filename} not found !")
            raise FileNotFoundError
        except KeyError:
            self.logger.error(f"csv file {csv_filename} has no CAS column !")
            raise KeyError
        except Exception as e:
            self.logger.error(f"csv file {csv_filename} has an error !")
            raise e



# pubchem data models

from pydantic import BaseModel
from typing import List, Optional, Dict

class Id(BaseModel):
    cid: int

class IdWrapper(BaseModel):
    id: Id

class Atoms(BaseModel):
    aid: List[int]
    element: List[int]

class Bonds(BaseModel):
    aid1: List[int]
    aid2: List[int]
    order: List[int]

class Style(BaseModel):
    annotation: List[int]
    aid1: List[int]
    aid2: List[int]

class Conformers(BaseModel):
    x: List[float]
    y: List[float]
    style: Style

class Coords(BaseModel):
    type: List[int]
    aid: List[int]
    conformers: List[Conformers]

class Urn(BaseModel):
    label: Optional[str] = None
    name: Optional[str] = None
    datatype: Optional[int]
    release: Optional[str]
    implementation: Optional[str] = None
    version: Optional[str] = None
    software: Optional[str] = None
    source: Optional[str] = None
    parameters: Optional[str] = None

class Value(BaseModel):
    sval: Optional[str] = None
    ival: Optional[int] = None
    fval: Optional[float] = None
    binary: Optional[str] = None

class Prop(BaseModel):
    urn: Urn
    value: Value

class Count(BaseModel):
    heavy_atom: int
    atom_chiral: int
    atom_chiral_def: int
    atom_chiral_undef: int
    bond_chiral: int
    bond_chiral_def: int
    bond_chiral_undef: int
    isotope_atom: int
    covalent_unit: int
    tautomers: int

class PCCompound(BaseModel):
    id: IdWrapper
    atoms: Atoms
    bonds: Bonds
    coords: List[Coords]
    charge: int
    props: List[Prop]
    count: Count

class PubChemRecord(BaseModel):
    PC_Compounds: List[PCCompound]


"""_____________________________________________________________________

:PROJECT: ChemDataReader

* Main module implementation *

:details:  Main module implementation.

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from typing import IO

import json
import asyncio
from aiohttp import ClientSession

from chemdatareader import chemdatareader_logger
from chemdatareader.collectors.pubchem_data_models import PubChemRecord

import chemdatareader.chemdatacollector_builder_interface as cdcbi

from chemdatareader.cached_request import CachedRequest


class PubChemCollector(cdcbi.ChemDataCollectorInterface):
    def __init__(self, access_code: str = None) -> None:
        """Implementation of the ChemDatReader Interface"""
        self.logger = chemdatareader_logger.get_logger(__name__)

        self.access_code = access_code

        self.PubChemURL = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/"

        self.cached_request = CachedRequest()

    def test_connection(self) -> str:
        """connection test module

        :param name: person to greet
        :type name: str
        """
        self.logger.debug(f"connection to : {self.PubChemURL}")
        self.logger.debug(f"access code : {self.access_code}")
        return self.PubChemURL

    async def get_compounds_by_name(
        self, name_list: list[str] = [], operation: str = "record", output_format: str = "JSON"
    ):
        """
        get all compounds of a list of names or CAS numbers
        """
        self.compound_list = []
        self.operation = operation
        self.output_format = output_format

        async with ClientSession() as session:
            tasks = []
            for name in name_list:
                # task = asyncio.ensure_future(self._get_compound_by_name(name, session))
                tasks.append(self._get_compound_by_name(name, session))
            await asyncio.gather(*tasks, return_exceptions=True)

    async def _get_compound_by_name(self, name: str = None, session: ClientSession = None):
        """
        get compound by name or CAS number
        """
        self.logger.debug(f"get compound by name / CAS : {name}")
        print(f"get compound by name / CAS --: {name}")

        response = await self.cached_request.get(
            session=session, url="/".join([self.PubChemURL, name, self.operation, self.output_format])
        )
        self.compound_list.append(response)

        if self.output_format == "JSON" and self.operation == "record":
            self.parse_record(response)

    def parse_record(self, record_json: str = None):
        """
        parse record
        """
        # self.logger.debug(f"parse record : {record}")
        # self.compound_list.append(response)
        record = PubChemRecord.model_validate_json(record_json)

        # add to record list

    def properties_to_dict(self, record: PubChemRecord):
        """
        convert properties to dict
        """
        for compound in record.PC_Compounds:
            properties_dict = {}
            for prop in compound.props:
                if prop.urn.name is not None:
                    key = prop.urn.name.replace(" ", "")
                key += prop.urn.label.replace(" ", "")
                value = prop.value.sval or prop.value.fval or prop.value.binary
                properties_dict[key] = value

            # Now you can use the properties_dict
            print(properties_dict)

            # convert properties to JSON

            properties_json = json.dumps(properties_dict)
            print(properties_json)


class PubChemServiceBuilder(cdcbi.ChemDatCollectorBuilderInterface):
    def __init__(self):
        self._instance = None

    def __call__(self, pubchem_client_key, pubchem_client_secret, **_ignored):
        if not self._instance:
            access_code = self.authorize(pubchem_client_key, pubchem_client_secret)
            print(f"authorised access_code : {access_code}")
            self._instance = PubChemCollector(access_code)
        return self._instance

    def authorize(self, key, secret):
        return "MY_PUBCHEM_ACCESS_CODE"


cdcbi.services.register_builder("pubchem", PubChemServiceBuilder())



#!/usr/bin/env python3
# vim:fileencoding=utf-8
"""_____________________________________________________________________

:PROJECT: ChemDataReader

* Main module command line interface *

:details:  Main module command line interface. 
           !!! Warning: it should have a diffent name than the package name.

.. note:: -
.. todo:: - 
________________________________________________________________________
"""

"""Main module implementation. !!! Warning: it should have a diffent name than the package name. """
"""Console script for chemdatareader."""

import logging
import argparse
import sys
import asyncio
from chemdatareader.chemdatareader_logger import get_logger
from chemdatareader import __version__

from chemdatareader.chemdatareader_impl import ChemDatReader

logging.basicConfig(level=logging.DEBUG)    

def parse_command_line():
    """ Looking for command line arguments"""

    description = "chemdatareader"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("_", nargs="*")

    parser.add_argument(
        "-f", "--csv-filename", action="store", help="filename of the csv file"
    )

    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)

    # add more arguments here

    return parser.parse_args()

async def main():
    """Console script for chemdatareader."""
        # or use logging.INFO (=20) or logging.ERROR (=30) for less output
    
    logger = get_logger(__name__)
    
    args = parse_command_line()
        
    if len(sys.argv) <= 2:
        logging.debug("no arguments provided !")

    #print("Arguments: " + str(args._))   
    
    cdr = ChemDatReader()
    if args.csv_filename:
        logger.debug(f"csv_filename : {args.csv_filename}")
        await cdr.read_cas_csv(args.csv_filename)
    
    return 0

def main_async():
    asyncio.run(main())

if __name__ == "__main__":
     sys.exit(main_async())  # pragma: no cover



#!/usr/bin/env python
"""Tests for `chemdatareader` package."""
# pylint: disable=redefined-outer-name
from chemdatareader import __version__
from chemdatareader.chemdatareader_interface import ChemDatReaderInterface
from chemdatareader.chemdatareader_impl import ChemDatReader
from chemdatareader.chemdatacollector_builder_interface import ChemDatCollectorBuilderInterface, ChemDataCollectorInterface
from chemdatareader.collectors.pubchem_collector import PubChemServiceBuilder, PubChemCollector


# def test_version():
#     """Sample pytest test function."""
#     assert __version__ == "0.0.1"

def test_ChemDataReaderInterface():
    """ testing the formal interface (ChemDataReaderInterface)
    """
    assert issubclass(ChemDatReader, ChemDatReaderInterface)
    

def test_ChemDataCollectorBuilderInterface():
    """ testing the formal interface (PubChemReaderInterface)
    """
    assert issubclass(PubChemServiceBuilder, ChemDatCollectorBuilderInterface)


def test_ChemDataCollectorInterface():
    """ testing the formal interface (PubChemReaderInterface)
    """
    assert issubclass(PubChemCollector, ChemDataCollectorInterface)

def test_PubChemReader():
    """ Testing PubChemService class
    """
    pchr = PubChemCollector()
    
    assert pchr.test_connection() == "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/"




import asyncio
import hashlib
import redis
import os
import pickle

from aiohttp import ClientSession

from chemdatareader.chemdatareader_logger import get_logger


def sha512_hash(text: str):
    return hashlib.sha256(text.encode("UTF-8")).hexdigest()


class CachedRequest:
    def __init__(
        self,
        redis_host: str = "localhost",
        redis_port: int = 6379,
        redis_db: str = "",
        redis_timeout: int = 604800,
        request_delay: float = 0.1,
    ) -> None:
        self.redis_host = os.getenv("REDIS_HOST", redis_host)
        self.redis_port = os.getenv("REDIS_PORT", redis_port)
        self.redis_user = os.getenv("REDIS_USER", "")
        self.redis_password = os.getenv("REDIS_PASSWORD", "")
        self.redis_db = os.getenv("REDIS_DB", redis_db)

        self.redis_timeout = os.getenv("REDIS_TIMEOUT", redis_timeout)
        self.request_delay = os.getenv("REQUEST_DELAY", request_delay)

        self.logger = get_logger(__name__)

        try:
            self.redis_srv = redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db, password=self.redis_password)
            #self.redis_srv = redis.Redis()
        except Exception as e:
            self.logger.error(f"redis connection error : {e}")

    async def get(self, url: str, session: ClientSession, **kwargs):
        if self.redis_srv is not None:
            redis_key = sha512_hash(url)

            try:
                value_redis_cached = self.redis_srv.get(redis_key)
                return pickle.loads(value_redis_cached)
            except Exception as e:
                self.logger.error(f"redis redis_key error : {e}")

                # redis value not available, fetch from url and store in redis
                try:
                    await asyncio.sleep(0.1)
                    value = await self.fetch_html(url, session, **kwargs)
                    self.redis_srv.set(redis_key, pickle.dumps(value), ex=self.redis_timeout)
                    return value
                except Exception as e:
                    self.logger.error(f"ERROR: caching to redis error : {e}")
                    raise e
        else:
            #redis server not available, fetch from url directly
            await asyncio.sleep(0.1)
            return await self.fetch_html(url, session, **kwargs)

    async def fetch_html(self, url: str, session: ClientSession, **kwargs) -> str:
        """GET request wrapper to fetch page HTML.

        kwargs are passed to `session.request()`.
        """
        await asyncio.sleep(self.request_delay)

        resp = await session.request(method="GET", url=url, **kwargs)
        resp.raise_for_status()
        # logger.info("Got response [%s] for URL: %s", resp.status, url)
        res_text = await resp.text()
        #self.logger.debug(f"Got response [{resp.status}] for URL: {url}\n: {res_text}")
        return res_text



import os
import re

from setuptools import setup, find_packages


REGEX_COMMENT = re.compile(r"[\s^]#(.*)")

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

dir_path = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(dir_path, "VERSION"), "r") as version_file:
    version = str(version_file.readline()).strip()


def parse_requirements(filename):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), filename)
    with open(filename, "rt") as filehandle:
        requirements = filehandle.readlines()[2:]
        return tuple(filter(None, (REGEX_COMMENT.sub("", line).strip() for line in requirements)))


setup(
    name="chemdatareader",
    version=version,
    packages=find_packages(),
    include_package_data=True,
)



"""_____________________________________________________________________

:PROJECT: ChemDataReader

* Main module formal interface. *

:details: In larger projects, formal interfaces are helping to define a trustable contract.
          Currently there are two commonly used approaches: 
          [ABCMetadata](https://docs.python.org/3/library/abc.html) or [Python Protocols](https://peps.python.org/pep-0544/)

       see also:
       ABC metaclass
         - https://realpython.com/python-interface/
         - https://dev.to/meseta/factories-abstract-base-classes-and-python-s-new-protocols-structural-subtyping-20bm

.. note:: -
.. todo:: - 
________________________________________________________________________
"""


from abc import ABCMeta, abstractmethod
import asyncio


class ChemDatReaderInterface(metaclass=ABCMeta):
    """ChemDatReader formal Interface
    TODO: test, if ABC baseclass is wor
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "read_csv") and callable(subclass.read_csv) or NotImplemented
        return NotImplemented

    @classmethod
    @abstractmethod
    async def read_substance_names_csv(self, csv_filename: str = None) -> str:
        """csv reader module

        :param name: person to greet
        :type name: str
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def get_compounds_by_name(self, name_list: list[str] = [], operation: str = None, output_format: str = None):
        """
        get all compounds of a list of names or CAS numbers
        """
        raise NotImplementedError



import logging

#_log_format = f"%(asctime)s - [%(levelname)s] - %(name)s - (%(filename)s).%(funcName)s(%(lineno)d) - %(message)s"
_log_format = f"%(asctime)s [%(levelname)s] (%(filename)s).%(funcName)s(%(lineno)d) - %(message)s"
_log_format_stream = f"%(asctime)s [%(levelname)s] (%(filename)s).%(funcName)s(%(lineno)d):\n\t%(message)s"

class CustomFilter(logging.Filter):
    COLOR = {
        "DEBUG": "GREEN",
        "INFO": "GREEN",
        "WARNING": "YELLOW",
        "ERROR": "RED",
        "CRITICAL": "RED",
    }

    def filter(self, record):
        record.color = CustomFilter.COLOR[record.levelname]
        return True
    
def get_file_handler():
    file_handler = logging.FileHandler("chem_data_reader.log")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(_log_format_stream))
    return file_handler

def get_stream_handler():
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(_log_format_stream))
    return stream_handler

def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(get_file_handler())

    logger.addHandler(get_stream_handler())
    logger.addFilter(CustomFilter())
    return logger
  
